# Example Use:
# from navlayers.manipulators.layerops import LayerOps
# from navlayers.core.layer import Layer

# demo = Layer()
# demo.load_file("C:\Users\attack\Downloads\layer.json")
# demo2 = Layer()
# demo2.load_input(Existing_Layer_String_Previously_Loaded)

# lo = LayerOps(score=lambda x: x[0] * x[1], name=lambda x: x[1],
#               desc=lambda x: "This is an list example")
# out_layer = lo.process([demo, demo2])
# out_layer.export_file("C:\demo_layer1.json")

# lo2 = LayerOps(score=lambda x: x['a'], color=lambda x: x['b'],
#                desc=lambda x: "This is a dict example")
# out_layer2 = lo2.process({'a': demo, 'b': demo2})
# dict_layer = out_layer2.get_dict()
# print(dict_layer)
# out_layer2.export_file("C:\demo_layer2.json")

import copy
from mitreattack.navlayers.core import Layer


class InvalidFormat(Exception):
    pass


class BadLambda(Exception):
    pass


class MismatchedDomain(Exception):
    pass


class LayerOps:
    def __init__(self, score=None, comment=None, enabled=None, colors=None,
                 metadata=None, name=None, desc=None, default_values=None):
        """
            Initialization - configures the object to handle processing
                based on user provided Lambdas
            :param score: lambda to calculate score
            :param comment: lambda to generate comments
            :param enabled: lambda to determine enabled status
            :param metadata: lambda to generate metadata
            :param name: new name to apply to the resulting layer
            :param desc: new description to apply to the resulting layer
            :param default_values: dictionary containing desired default
                values for missing data element values
        """
        self._score = score
        self._comment = comment
        self._enabled = enabled
        self._colors = colors
        self._metadata = metadata
        self._name = name
        self._desc = desc
        self._default_values = {
            "comment": "",
            "enabled": True,
            "color": "#ffffff",
            "score": 1,
            "metadata": []
        }
        if default_values is not None:
            for entry in default_values:
                self._default_values[entry] = default_values[entry]

    def process(self, data, default_values=None):
        """
            takes a list or dict of Layer objects, and processes them
            :param data: A dict or list of Layer objects.
            :param default_values: dictionary containing desired default values for
                missing data element values
            :raises InvalidFormat: An error indicating that the layer data
                wasn't provided in a list or dict
        """
        if isinstance(data, dict):
            temp = {}
            for entry in data.keys():
                temp[entry] = data[entry].layer.techniques
            self.mode = 'dict'
            out = self._merge_to_template(data, key=list(data.keys())[0])
        elif isinstance(data, list):
            temp = []
            for entry in data:
                temp.append(entry.layer.techniques)
            self.mode = 'list'
            out = self._merge_to_template(data)
        else:
            raise InvalidFormat
        da = temp
        corpus = self._build_template(temp)

        defaults = self._default_values
        if default_values is not None:
            for entry in default_values:
                defaults[entry] = default_values[entry]

        return self._compute(data, da, corpus, out, defaults)

    def _compute(self, data, da, corpus, out, defaults):
        """
            INTERNAL: Applies the configured lambda to the dataset
            :param data: the dataset being processed
            :param da: extracted techniques from the dataset, sorted by
                dataset format
            :param corpus: master list of combined techniques and
                technique data
            :param out: baseline template for the new layer
            :param defaults: default values each technique should use if a
                field is missing
            :returns: a Layer object representing the resultant layer
        """
        composite = copy.deepcopy(corpus)
        if self._score is not None:
            for entry in composite:
                entry['score'] = self._applyOperation(da, entry, 'score',
                                                      self._score, defaults)

        if self._comment is not None:
            for entry in composite:
                entry['comment'] = self._applyOperation(da, entry, 'comment',
                                                        self._comment, defaults)

        if self._enabled is not None:
            for entry in composite:
                entry['enabled'] = self._applyOperation(da, entry, 'enabled',
                                                        self._enabled, defaults)

        if self._colors is not None:
            for entry in composite:
                entry['color'] = self._applyOperation(da, entry, 'color',
                                                      self._colors, defaults)

        if self._metadata is not None:
            for entry in composite:
                entry['metadata'] = self._applyOperation(da, entry, 'metadata',
                                                         self._metadata,
                                                         defaults)

        processed = copy.deepcopy(out)
        processed['techniques'] = composite
        if self._name is not None:
            processed['name'] = self._applyOperation(data, None, 'name',
                                                     self._name, defaults,
                                                     glob='name')

        if self._desc is not None:
            processed['description'] = self._applyOperation(data, None,
                                                            'description',
                                                            self._desc,
                                                            defaults,
                                                            glob='description')

        return Layer(processed)

    def _merge_to_template(self, data, key=0):
        """
            INTERNAL: merges initial layer files in either dict or list form
                into a single template. Defaults to the first entry in the
                case of difference in metadata.
            :param key: the key referencing the first entry to default to
            :raises MismatchedDomain: An error indicating that the layers
                came from different domains
        """
        out = {}
        dict_map = []
        collide = []
        if self.mode == 'dict':
            for x in data.keys():
                dict_map.append(x)
                collide.append(data[x])
        else:
            for x in data:
                collide.append(x)
        key_space = data[key].layer._enumerate()
        _raw = data[key].layer.get_dict()
        for entry in key_space:
            if entry != 'techniques':
                standard = _raw[entry]
                if any(y != standard for y in
                       [getattr(x.layer, entry) for x in collide]):
                    if entry == 'domain':
                        print('FATAL ERROR! Layer mis-match on domain. '
                              'Exiting.')
                        raise MismatchedDomain
                    print('Warning! Layer mis-match detected for {}. '
                          'Defaulting to {}\'s value'.format(entry, key))
                out[entry] = standard
        return out

    def _build_template(self, data):
        """
            INTERNAL: builds a base template by combining available technique
                listings from each layer
            :param data: the raw ingested technique data (list or dict)
        """
        if self.mode == 'list':
            return self._template(data)
        elif self.mode == 'dict':
            temp = {}
            t2 = []
            for key in data:
                temp[key] = self._template([data[key]])
            for key in temp:
                for elm in temp[key]:
                    if not any(elm['techniqueID'] == x['techniqueID']
                               and elm['tactic'] == x['tactic']
                               for x in t2):
                        t2.append(elm)
                    else:
                        [x.update(elm)
                         if x['techniqueID'] == elm['techniqueID']
                            and elm['tactic'] == x['tactic']
                         else x for x in t2]
            return t2

    @staticmethod
    def _template(data):
        """
            INTERNAL: creates a template technique entry for a given listing
                of techniques
            :param data: a single layer's technique data
            :returns: a list of technique templates
        """
        temp = []
        for entry in data:
            temp.append([{"techniqueID": x.techniqueID, "tactic": x.tactic}
                         if x.tactic else {"techniqueID": x.techniqueID}
                         for x in entry])
        complete = []
        for entry in temp:
            for val in entry:
                if val in complete:
                    continue
                complete.append(val)
        return complete

    def _grabList(self, search, collection):
        """
            INTERNAL: generates a list containing all values for a given key
                across the collection
            :param search: the key to search for
            :param collection: the data collection to search
            :returns: a list of values for that key across the collection
        """
        temp = []
        for x in collection:
            temp.append(self._grabElement(search, x))
        return temp

    def _grabDict(self, search, collection):
        """
            INTERNAL: generates a dictionary containing all values for a given
                key across the collection
            :param search: the key to search for
            :param collection: the data collection to search
            :returns: a dict of values for that key across the collection
        """
        temp = {}
        for key in collection:
            temp[key] = self._grabElement(search, collection[key])
        return temp

    def _grabElement(self, search, listing):
        """
            INTERNAL: returns a matching element in the listing for the
                search key
            :param search: the key to search for
            :param listing: the data element to search
            :returns: the found value, or an empty dict
        """
        val = [x for x in listing if self._inDict(search, x)]
        if len(val) > 0:
            return val[0]
        return {}

    @staticmethod
    def _inDict(search, complete):
        """
            INTERNAL: returns bool of whether or not the key searched
                for can be found across the dataset corpus
            :param search: the key to search for
            :param complete: the data set to search for
            :returns: true/false
        """
        comp_list = complete.get_dict().items()
        search_terms = search.items()
        return all(elm in comp_list for elm in search_terms)

    def _applyOperation(self, corpus, element, name, lda, defaults, glob=None):
        """
            INTERNAL: applies a lambda expression to the dataset
            :param corpus: the dataset
            :param element: the template file to fill out
            :param name: the name of the field being processed
            :param lda: the lambda expression to apply
            :param defaults: any default values should the field not be found
                in a dataset entry
            :raises BadLambda: error denoting that an error has occurred when
                running the lambda function
            :returns: lambda output
        """
        if self.mode == 'list':
            if glob:
                listing = [getattr(x.layer, glob) for x in corpus]
                listing = [{name: x} for x in listing]
            else:
                te = dict(techniqueID=element['techniqueID'])
                if 'tactic' in element:
                    te['tactic'] = element['tactic']
                listing = self._grabList(te, corpus)
                listing = [x.get_dict() if not isinstance(x, dict)
                           else dict() for x in listing]
            values = []
            for x in listing:
                if name in x:
                    values.append(x[name])
                else:
                    if defaults is not None:
                        if name in defaults:
                            values.append(defaults[name])
                            continue
                    values.append(self._default_values[name])
        else:
            values = {}
            if glob:
                listing = {}
                for k in corpus.keys():
                    listing[k] = {glob: getattr(corpus[k].layer, glob)}
            else:
                te = dict(techniqueID=element['techniqueID'])
                if 'tactic' in element:
                    te['tactic'] = element['tactic']
                temp = self._grabDict(te, corpus)
                listing = {}
                for k in temp.keys():
                    if temp[k] != {}:
                        listing[k] = temp[k].get_dict()
                    else:
                        listing[k] = {}
            for elm in listing:
                if name in listing[elm]:
                    values[elm] = listing[elm][name]
                else:
                    if defaults is not None:
                        if name in defaults:
                            values[elm] = defaults[name]
                            continue
                    values[elm] = self._default_values[name]
        try:
            return lda(values)
        except IndexError and KeyError:
            print('Unable to continue, lambda targeting "{}" could not operate'
                  ' correctly on {}. Maybe the field is missing?'
                  .format(name, element))
            print('[RAW] Extracted matching elements: {}'.format(listing))
            raise BadLambda



from .layer import Layer
from .exceptions import *
from .filter import Filter
from .gradient import Gradient
from .layerobj import _LayerObj
from .layout import Layout
from .legenditem import LegendItem
from .metadata import Metadata, MetaDiv
from .technique import Technique
from .versions import Versions
from .helpers import *
from .objlink import Link, LinkDiv



from mitreattack.navlayers.core.exceptions import typeChecker, categoryChecker, UNSETVALUE, BadInput

defaults = dict(layer="4.3", navigator="4.5.5")

class Versions:
    def __init__(self, layer=defaults['layer'], attack=UNSETVALUE, navigator=defaults['navigator']):
        """
            Initialization - Creates a v4 Versions object

            :param layer: The layer version
            :param attack: The attack version
            :param navigator: The navigator version
        """
        self.layer = layer
        self.__attack = attack
        self.navigator = navigator

    @property
    def attack(self):
        if self.__attack != UNSETVALUE:
            return self.__attack
        else:
            return '4.x'

    @attack.setter
    def attack(self, attack):
        typeChecker(type(self).__name__, attack, str, "attack")
        self.__attack = attack

    @property
    def navigator(self):
        return self.__navigator

    @navigator.setter
    def navigator(self, navigator):
        typeChecker(type(self).__name__, navigator, str, "navigator")
        if not navigator.startswith('4.'):
            print(f'[WARNING] - unrecognized navigator version {navigator}. Defaulting to the 4.X schema, '
                  f'this may result in unexpected behavior.')
            navigator = defaults['navigator']
        self.__navigator = navigator

    @property
    def layer(self):
        return self.__layer

    @layer.setter
    def layer(self, layer):
        typeChecker(type(self).__name__, layer, str, "layer")
        try:
            categoryChecker(type(self).__name__, layer, ["3.0", "4.0", "4.1", "4.2", "4.3"], "layer version")
        except BadInput:
            print(f'[WARNING] - unrecognized layer version {layer}. Defaulting to the 4.3 schema, this may result in '
                  f'unexpected behavior.')
        if layer in ["3.0", "4.0", "4.1", "4.2"]:
            print(f'[NOTICE] - Forcibly upgrading version from {layer} to 4.3.')
            layer = "4.3"
        self.__layer = layer

    def get_dict(self):
        """
            Converts the currently loaded data into a dict
            :returns: A dict representation of the local Versions object
        """
        temp = dict()
        listing = vars(self)
        for entry in listing:
            if listing[entry] != UNSETVALUE:
                subname = entry.split('__')[-1]
                temp[subname] = listing[entry]
        return temp



from copy import deepcopy
from openpyxl.comments import Comment
from openpyxl.styles import PatternFill, Font
import colorsys

from mitreattack.navlayers.core import Layer, Gradient
from mitreattack.navlayers.exporters import ExcelTemplates


class ToExcel:
    def __init__(self, domain='enterprise', source='taxii', resource=None):
        """
            Sets up exporting system, builds underlying matrix

            :param source: Source to generate the matrix from, one of (taxii, local, or remote)
            :param resource: Optional string path to local cache of stix data (local) or url of workbench to reach out
                                to (remote)

        """
        self.domain = domain
        self.raw_handle = ExcelTemplates(domain=domain, source=source, resource=resource)

    def to_xlsx(self, layerInit, filepath="layer.xlsx"):
        """
            Exports a layer file to the excel format as the file specified

            :param layerInit: A layer to initialize the instance with
            :param filepath: The location to write the excel file to
        """

        if not isinstance(layerInit, Layer):
            raise TypeError

        layer = deepcopy(layerInit)

        if self.domain not in layer.layer.domain:
            raise ValueError(f"layer domain ({layer.layer.domain}) does not match exporter domain ({self.domain})")

        included_subs = []
        if layer.layer.techniques:
            for entry in layer.layer.techniques:
                if entry.showSubtechniques:
                    if entry.tactic:
                        included_subs.append((entry.techniqueID, entry.tactic))
                    else:
                        included_subs.append((entry.techniqueID, False))

        excluded = []
        if layer.layer.hideDisabled:
            for entry in layer.layer.techniques:
                if entry.enabled is False:
                    if entry.tactic:
                        excluded.append((entry.techniqueID, entry.tactic))
                    else:
                        excluded.append((entry.techniqueID, False))
        scores = []
        if layer.layer.techniques:
            for entry in layer.layer.techniques:
                if entry.score is not None:
                    if entry.tactic:
                        scores.append((entry.techniqueID, entry.tactic, entry.score))
                    else:
                        scores.append((entry.techniqueID, False, entry.score))
        sName = True
        sID = False
        sort = 0
        if layer.layer.layout:
            sName = layer.layer.layout.showName
            sID = layer.layer.layout.showID
        if layer.layer.sorting:
            sort = layer.layer.sorting
        raw_template = self.raw_handle.export(showName=sName, showID=sID, filters=layer.layer.filters, sort=sort,
                                              scores=scores, subtechs=included_subs, exclude=excluded)
        sheet_obj = raw_template.active
        sheet_obj.title = layer.layer.name
        # v4.2 - do aggregate adjustments
        if layer.layer.layout:
            if layer.layer.layout.showAggregateScores:
                for tac_column in self.raw_handle.codex:
                    short_hand = self.raw_handle.h.convert(tac_column.tactic.name)
                    for x in tac_column.techniques:
                        x_score = [y for y in scores if (y[0] == x.id and (y[1] == short_hand or y[1] is None))]
                        if len(x_score):
                            x.score = x_score[0][2]
                        subs = tac_column.subtechniques.get(x.id, [])
                        for sub_score in subs:
                            subtech_score = [y for y in scores if (y[0] == sub_score.id and
                                                                   (y[1] == short_hand or y[1] is None))]
                            if len(subtech_score):
                                sub_score.score = subtech_score[0][2]
                        mod = layer.layer.layout.compute_aggregate(x, subs)
                        patch_target = [y for y in layer.layer.techniques if (y.techniqueID == x.id and
                                                                              (y.tactic == short_hand or
                                                                               y.tactic is None))]
                        if len(patch_target):
                            patch_target[0].score = mod
                        elif mod:
                            print("[WARNING] - Aggregate calculated for a technique that doesn't seem to exist...")

        # verify gradient information
        safe_gradient = layer.layer.gradient
        if not safe_gradient:
            safe_gradient = Gradient(colors=["#ff6666", "#ffe766", "#8ec843"], minValue=1, maxValue=100)

        for tech in layer.layer.techniques:
            p_tactic = None
            if tech.tactic:
                p_tactic = tech.tactic
            coords = self.raw_handle.retrieve_coords(tech.techniqueID, p_tactic)
            if coords == [] or coords == 'HIDDEN':
                tac = p_tactic
                if tac is None:
                    tac = "(none)"
                if coords:
                    print('WARNING! Technique/Tactic ' + tech.techniqueID + '/' + tac +
                          ' does not appear to exist in the loaded matrix. Skipping...')
                else:
                    parents = [x for x in layer.layer.techniques if x.techniqueID == tech.techniqueID.split('.')[0]]
                    if tech.tactic:
                        parents = [x for x in parents if x.tactic == tech.tactic]
                    if all([True if not x.showSubtechniques else False for x in parents]):
                        print('NOTE! Technique/Tactic ' + tech.techniqueID + '/' + tac + ' does not appear '
                              'to be visible in the matrix. Its parent appears to be hiding it.')
                    else:
                        print('WARNING! Technique/Tactic ' + tech.techniqueID + '/' + tac + ' seems malformed. '
                              'Skipping...')
                    continue
            for location in coords:
                cell = sheet_obj.cell(row=location[0], column=location[1])
                if tech.comment:
                    cell.comment = Comment(tech.comment, 'ATT&CK Scripts Exporter')

                if tech.enabled is False:
                    if layer.layer.hideDisabled:
                        pass
                    else:
                        grayed_out = Font(color='909090')
                        cell.font = grayed_out
                        continue
                if tech.color:
                    c_color = PatternFill(fill_type='solid', start_color=tech.color.upper()[1:])
                    cell.fill = c_color
                    continue
                if tech.score is not None:
                    tscore = tech.score
                    comp_color = safe_gradient.compute_color(tscore)
                    c_color = PatternFill(fill_type='solid', start_color=comp_color.upper()[1:])
                    cell.fill = c_color
                    RGB = tuple(int(comp_color.upper()[1:][i:i+2], 16) for i in (0, 2, 4))
                    hls = colorsys.rgb_to_hls(RGB[0], RGB[1], RGB[2])
                    if hls[1] < 127.5:
                        white = Font(color='FFFFFF')
                        cell.font = white
        raw_template.save(filepath)



import argparse
import datetime
import json
import os
from itertools import chain
from pathlib import Path
from typing import List

import markdown
import requests
import urllib3
from dateutil import parser as dateparser
from loguru import logger
from stix2 import Filter, MemoryStore, TAXIICollectionSource
from taxii2client.v20 import Collection
from tqdm import tqdm

# helper maps
domainToDomainLabel = {"enterprise-attack": "Enterprise", "mobile-attack": "Mobile"}
domainToTaxiiCollectionId = {
    "enterprise-attack": "95ecc380-afe9-11e4-9b6c-751b66dd541e",
    "mobile-attack": "2f669986-b40b-4423-b720-4396ca6a462b",
}
# stix filters for querying for each type of data
attackTypeToStixFilter = {
    "technique": [Filter("type", "=", "attack-pattern")],
    "software": [Filter("type", "=", "malware"), Filter("type", "=", "tool")],
    "group": [Filter("type", "=", "intrusion-set")],
    "mitigation": [Filter("type", "=", "course-of-action")],
    "datasource": [
        Filter("type", "=", "x-mitre-data-source"),
        Filter("type", "=", "x-mitre-data-component"),
    ],
    "datasource-only": [Filter("type", "=", "x-mitre-data-source")],
}
# ATT&CK type to Title
attackTypeToTitle = {
    "technique": "Techniques",
    "malware": "Malware",
    "software": "Software",
    "group": "Groups",
    "mitigation": "Mitigations",
    "datasource": "Data Sources and/or Components",
}
# ATT&CK type to section name
attackTypeToSectionName = {
    "technique": "Technique",
    "malware": "Malware",
    "software": "Software",
    "group": "Group",
    "mitigation": "Mitigation",
    "datasource": "Data Source and/or Component",
}
# how we want to format headers for each section
sectionNameToSectionHeaders = {
    "additions": "New {obj_type}",
    "changes": "{obj_type} changes",
    "minor_changes": "Minor {obj_type} changes",
    "deprecations": "{obj_type} deprecations",
    "revocations": "{obj_type} revocations",
    "deletions": "{obj_type} deletions",
    "unchanged": "Unchanged {obj_type}",
}
# color key for layers
statusToColor = {
    "additions": "#a1d99b",
    "changes": "#fcf3a2",
    "minor_changes": "#c7c4e0",
    "deletions": "#ff00e1",  # this will probably never show up but just in case
    "revocations": "#ff9000",
    "deprecations": "#ff6363",
    "unchanged": "#ffffff",
}
# explanation of modification types to data objects for legend in layer files
statusDescriptions = {
    "additions": "objects which are present in the new data and not the old",
    "changes": "objects which have a newer version number in the new data compared to the old",
    "minor_changes": "objects which have a newer last edit date in the new data than in the old, but the same version number",
    "revocations": "objects which are revoked in the new data but not in the old",
    "deprecations": "objects which are deprecated in the new data but not in the old",
    "deletions": "objects which are present in the old data but not the new",
    "unchanged": "objects which did not change between the two versions",
}
date = datetime.datetime.today()
this_month = date.strftime("%B_%Y")
layer_defaults = [
    os.path.join("output", f"{this_month}_Updates_Enterprise.json"),
    os.path.join("output", f"{this_month}_Updates_Mobile.json"),
    os.path.join("output", f"{this_month}_Updates_Pre.json"),
]
md_default = os.path.join("output", f"updates-{this_month.lower()}.md")
json_default = os.path.join("output", f"updates-{this_month.lower()}.json")


class DiffStix(object):
    """Utilities for detecting and summarizing differences between two versions of the ATT&CK content."""

    def __init__(
        self,
        domains=["enterprise-attack", "mobile-attack"],
        layers=None,
        markdown=None,
        minor_changes=False,
        unchanged=False,
        new="new",
        old="old",
        show_key=False,
        site_prefix="",
        types=["technique", "software", "group", "mitigation", "datasource"],
        use_taxii=False,
        use_mitre_cti=False,
        verbose=False,
        include_contributors=False,
        release_contributors={},
    ):
        """Construct a new DiffStix object.

        params:
            domains: list of domains to parse, e.g. enterprise-attack, mobile-attack
            layers: array of output filenames for layer files, e.g. ['enterprise.json', 'mobile.json', 'pre.json']
            markdown: output filename for markdown content to be written to
            minor_changes: if true, also report minor changes section (changes which didn't increment version)
            new: directory to load for new stix version
            old: directory to load for old stix version
            show_key: if true, output key to markdown file
            site_prefix: prefix links in markdown output
            types: which types of objects to report on, e.g technique, software
            verbose: if true, print progress bar and status messages to stdout
        """
        self.domains = domains
        self.layers = layers
        self.markdown = markdown
        self.minor_changes = minor_changes
        self.unchanged = unchanged
        self.new = new
        self.old = old
        self.show_key = show_key
        self.site_prefix = site_prefix
        self.types = types
        self.use_taxii = use_taxii
        self.use_mitre_cti = use_mitre_cti
        self.verbose = verbose
        self.include_contributors = include_contributors
        # will hold information of contributors of the new release {... {"contributor_credit/name_as_key": counter]} ...}
        self.release_contributors = {}

        # data gets load into here in the load() function. All other functionalities rely on this data structure
        self.data = {
            # {
            #   technique: {
            #     enterprise-attack: {
            #       additions: [],
            #       deletions: [],
            #       changes: [],
            #       minor_changes: [],
            #       revocations: [],
            #       deprecations: [],
            #       unchanged: [],
            #     },
            #     mobile-attack: {...},
            #   },
            #   software: {...},
            # }
        }

        # stixID to object name
        self.stixIDToName = {}

        # all subtechnique-of relationships in the new/old data
        self.new_subtechnique_of_rels = []
        self.old_subtechnique_of_rels = []

        # all data components in the new/old data
        self.new_datacomponents = []
        self.old_datacomponents = []

        # stixID => technique for every technique in the new/old data
        self.new_id_to_technique = {}
        self.old_id_to_technique = {}

        # stixID => data source for every data source in the new/old data
        self.new_id_to_datasource = {}
        self.old_id_to_datasource = {}

        # build the bove data structures
        self.load_data()

        logger.info("removing duplicate relationships")
        self.new_subtechnique_of_rels = [
            i
            for n, i in enumerate(self.new_subtechnique_of_rels)
            if i not in self.new_subtechnique_of_rels[n + 1 :]
        ]
        self.old_subtechnique_of_rels = [
            i
            for n, i in enumerate(self.old_subtechnique_of_rels)
            if i not in self.old_subtechnique_of_rels[n + 1 :]
        ]

        logger.info("removing duplicate data components")
        self.new_datacomponents = [
            i
            for n, i in enumerate(self.new_datacomponents)
            if i not in self.new_datacomponents[n + 1 :]
        ]
        self.old_datacomponents = [
            i
            for n, i in enumerate(self.old_datacomponents)
            if i not in self.old_datacomponents[n + 1 :]
        ]

    def getUrlFromStix(self, datum, is_subtechnique=False):
        """
        Parse the website url from a stix object.
        """
        if datum.get("external_references"):
            url = datum["external_references"][0]["url"]
            split_url = url.split("/")
            splitfrom = -3 if is_subtechnique else -2
            link = "/".join(split_url[splitfrom:])
            return link
        return None

    def getDataComponentUrl(self, datasource, datacomponent):
        """Create url of data component with parent data source"""
        return f"{self.getUrlFromStix(datasource)}/#{'%20'.join(datacomponent['name'].split(' '))}"

    def deep_copy_stix(self, objects):
        """Transform stix to dict and deep copy the dict."""
        result = []
        for obj in objects:
            obj = dict(obj)
            if "external_references" in obj:
                for i in range(len(obj["external_references"])):
                    obj["external_references"][i] = dict(obj["external_references"][i])
            if "kill_chain_phases" in obj:
                for i in range(len(obj["kill_chain_phases"])):
                    obj["kill_chain_phases"][i] = dict(obj["kill_chain_phases"][i])
            if "modified" in obj:
                obj["modified"] = str(obj["modified"])
            if "definition" in obj:
                obj["definition"] = dict(obj["definition"])
            obj["created"] = str(obj["created"])
            result.append(obj)
        return result

    # load data into data structure
    def load_data(self):
        """Load data from files into data dict."""
        pbar = tqdm(
            total=len(self.types) * len(self.domains),
            desc="loading data",
            bar_format="{l_bar}{bar}| [{elapsed}<{remaining}, {rate_fmt}{postfix}]",
        )
        for obj_type in self.types:
            for domain in self.domains:

                def load_datastore(data_store):
                    """Handle data loaded from either a directory or the TAXII server"""
                    raw_data = list(
                        chain.from_iterable(
                            data_store.query(f)
                            for f in attackTypeToStixFilter[obj_type]
                        )
                    )
                    raw_data = self.deep_copy_stix(raw_data)
                    id_to_obj = {item["id"]: item for item in raw_data}

                    return {
                        "id_to_obj": id_to_obj,
                        "keys": set(id_to_obj.keys()),
                        "data_store": data_store,
                    }

                def parse_subtechniques(data_store, new=False):
                    """Parse dataStore sub-technique-of relationships"""
                    if new:
                        for technique in list(
                            data_store.query(attackTypeToStixFilter["technique"])
                        ):
                            self.new_id_to_technique[technique["id"]] = technique
                        self.new_subtechnique_of_rels += list(
                            data_store.query(
                                [
                                    Filter("type", "=", "relationship"),
                                    Filter("relationship_type", "=", "subtechnique-of"),
                                ]
                            )
                        )
                    else:
                        for technique in list(
                            data_store.query(attackTypeToStixFilter["technique"])
                        ):
                            self.old_id_to_technique[technique["id"]] = technique
                        self.old_subtechnique_of_rels += list(
                            data_store.query(
                                [
                                    Filter("type", "=", "relationship"),
                                    Filter("relationship_type", "=", "subtechnique-of"),
                                ]
                            )
                        )

                def parse_datacomponents(data_store, new=False):
                    """Parse dataStore x-mitre-data-components"""
                    if new:
                        for datasource in list(
                            data_store.query(attackTypeToStixFilter["datasource-only"])
                        ):
                            self.new_id_to_datasource[datasource["id"]] = datasource
                        self.new_datacomponents += list(
                            data_store.query(
                                [Filter("type", "=", "x-mitre-data-component")]
                            )
                        )
                    else:
                        for datasource in list(
                            data_store.query(attackTypeToStixFilter["datasource-only"])
                        ):
                            self.old_id_to_datasource[datasource["id"]] = datasource
                        self.old_datacomponents += list(
                            data_store.query(
                                [Filter("type", "=", "x-mitre-data-component")]
                            )
                        )

                def update_contributors(old_object, new_object):
                    """Update contributors list if new object has contributors"""
                    if new_object.get("x_mitre_contributors"):
                        new_object_contributors = set(
                            new_object["x_mitre_contributors"]
                        )

                        # Check if old objects had contributors
                        if old_object is None or not old_object.get(
                            "x_mitre_contributors"
                        ):
                            old_object_contributors = set()
                        else:
                            old_object_contributors = set(
                                old_object["x_mitre_contributors"]
                            )

                        # Remove old contributors from showing up
                        # if contributors are the same the result will be empty
                        new_contributors = (
                            new_object_contributors - old_object_contributors
                        )

                        # Update counter of contributor to track contributions
                        for new_contributor in new_contributors:
                            if self.release_contributors.get(new_contributor):
                                self.release_contributors[new_contributor] += 1
                            else:
                                self.release_contributors[new_contributor] = 1

                def load_dir(dir, new=False):
                    """Load data from directory according to domain"""
                    data_store = MemoryStore()
                    datafile = os.path.join(dir, domain + ".json")
                    data_store.load_from_file(datafile)
                    parse_subtechniques(data_store, new)
                    parse_datacomponents(data_store, new)
                    return load_datastore(data_store)

                def load_taxii(new=False):
                    """Load data from TAXII server according to domain"""
                    collection = Collection(
                        "https://cti-taxii.mitre.org/stix/collections/"
                        + domainToTaxiiCollectionId[domain]
                    )
                    data_store = TAXIICollectionSource(collection)
                    parse_subtechniques(data_store, new)
                    parse_datacomponents(data_store, new)
                    return load_datastore(data_store)

                def load_mitre_cti(new=False):
                    """Load data from MITRE CTI repo according to domain"""
                    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
                    stix_json = requests.get(
                        f"https://raw.githubusercontent.com/mitre/cti/master/{domain}/{domain}.json",
                        verify=False,
                    )
                    if stix_json.status_code == 200:
                        stix_json = stix_json.json()
                        data_store = MemoryStore(stix_data=stix_json["objects"])
                        parse_subtechniques(data_store, new)
                        parse_datacomponents(data_store, new)
                        return load_datastore(data_store)
                    exit(f"\n{domain} stix bundle download was unsuccessful")

                logger.debug(f"Loading: [{domain:17}]/{obj_type}")

                if self.use_taxii:
                    old = load_taxii(False)
                elif self.use_mitre_cti:
                    old = load_mitre_cti(False)
                else:
                    old = load_dir(self.old, False)
                new = load_dir(self.new, True)

                intersection = old["keys"] & new["keys"]
                additions = new["keys"] - old["keys"]
                deletions = old["keys"] - new["keys"]

                # sets to store the ids of objects for each section
                changes = set()
                minor_changes = set()
                revocations = set()
                deprecations = set()
                unchanged = set()

                # find changes, revocations and deprecations
                for key in intersection:

                    # find revoked objects
                    if (
                        "revoked" in new["id_to_obj"][key]
                        and new["id_to_obj"][key]["revoked"]
                    ):
                        # if it was previously revoked, it's not a change
                        if (
                            not "revoked" in old["id_to_obj"][key]
                            or not old["id_to_obj"][key]["revoked"]
                        ):
                            # store the revoking object
                            revoked_by_key = new["data_store"].query(
                                [
                                    Filter("type", "=", "relationship"),
                                    Filter("relationship_type", "=", "revoked-by"),
                                    Filter("source_ref", "=", key),
                                ]
                            )
                            if len(revoked_by_key) == 0:
                                logger.error(
                                    f"[{key}] revoked object has no revoked-by relationship"
                                )
                                continue
                            else:
                                revoked_by_key = revoked_by_key[0]["target_ref"]

                            new["id_to_obj"][key]["revoked_by"] = new["id_to_obj"][
                                revoked_by_key
                            ]

                            revocations.add(key)
                        # else it was already revoked, and not a change; do nothing with it

                    # find deprecated objects
                    elif (
                        "x_mitre_deprecated" in new["id_to_obj"][key]
                        and new["id_to_obj"][key]["x_mitre_deprecated"]
                    ):
                        # if previously deprecated, not a change
                        if not "x_mitre_deprecated" in old["id_to_obj"][key]:
                            deprecations.add(key)

                    # find all other changed objects
                    else:
                        # try getting version numbers; should only lack version numbers if something has gone
                        # horribly wrong or a revoked object has slipped through
                        try:
                            old_version = float(
                                old["id_to_obj"][key]["x_mitre_version"]
                            )
                        except ValueError:
                            logger.error(
                                f"ERROR: cannot get old version for object: {key}"
                            )

                        try:
                            new_version = float(
                                new["id_to_obj"][key]["x_mitre_version"]
                            )
                        except ValueError:
                            logger.error(
                                f"ERROR: cannot get new version for object: {key}"
                            )

                        # Verify if there are new contributors on the object
                        update_contributors(
                            old["id_to_obj"][key], new["id_to_obj"][key]
                        )

                        # check for changes
                        if new_version > old_version:
                            # an update has occurred to this object
                            changes.add(key)
                        else:
                            # check for minor change; modification date increased but not version
                            old_date = dateparser.parse(
                                old["id_to_obj"][key]["modified"]
                            )
                            new_date = dateparser.parse(
                                new["id_to_obj"][key]["modified"]
                            )
                            if new_date > old_date:
                                minor_changes.add(key)
                            else:
                                unchanged.add(key)

                # Add contributions from additions
                for key in additions:
                    update_contributors(None, new["id_to_obj"][key])

                # set data
                if obj_type not in self.data:
                    self.data[obj_type] = {}
                self.data[obj_type][domain] = {
                    "additions": [new["id_to_obj"][key] for key in additions],
                    "changes": [new["id_to_obj"][key] for key in changes],
                }
                # only create minor_changes data if we want to display it later
                if self.minor_changes:
                    self.data[obj_type][domain]["minor_changes"] = [
                        new["id_to_obj"][key] for key in minor_changes
                    ]

                # ditto for unchanged
                if self.unchanged:
                    self.data[obj_type][domain]["unchanged"] = [
                        new["id_to_obj"][key] for key in unchanged
                    ]

                self.data[obj_type][domain]["revocations"] = [
                    new["id_to_obj"][key] for key in revocations
                ]
                self.data[obj_type][domain]["deprecations"] = [
                    new["id_to_obj"][key] for key in deprecations
                ]

                # only show deletions if objects were deleted
                if len(deletions) > 0:
                    self.data[obj_type][domain]["deletions"] = [
                        old["id_to_obj"][key] for key in deletions
                    ]

                logger.debug(f"Loaded:  [{domain:17}]/{obj_type}")
                pbar.update(1)
        pbar.close()

    def get_md_key(self):
        """Create string describing each type of difference (change, addition, etc).

        Used in get_markdown_string.

        Includes minor changes if the DiffStix instance was instantiated with the minor_changes argument.

        Includes deletions if the changes include deletions.
        """

        have_deletions = False
        for types in self.data.keys():
            for domain in self.data[types].keys():
                if "deletions" in self.data[types][domain].keys():
                    have_deletions = True

        key = "#### Key\n\n"
        key += (
            "* New objects: " + statusDescriptions["additions"] + "\n"
            "* Object changes: " + statusDescriptions["changes"] + "\n"
        )
        if self.minor_changes:
            key += (
                "* Minor object changes: " + statusDescriptions["minor_changes"] + "\n"
            )
        if self.unchanged:
            key += "* Unchanged objects: " + statusDescriptions["unchanged"] + "\n"
        key += (
            "* Object revocations: " + statusDescriptions["revocations"] + "\n"
            "* Object deprecations: " + statusDescriptions["deprecations"]
        )
        if have_deletions:
            key += "\n" + "* Object deletions: " + statusDescriptions["deletions"]
        return key

    def has_subtechniques(self, sdo, new=False):
        """Return true or false depending on whether the SDO has sub-techniques.

        new determines whether to parse from the new or old data"""
        if new:
            return (
                len(
                    list(
                        filter(
                            lambda rel: rel["target_ref"] == sdo["id"],
                            self.new_subtechnique_of_rels,
                        )
                    )
                )
                > 0
            )
        else:
            return (
                len(
                    list(
                        filter(
                            lambda rel: rel["target_ref"] == sdo["id"],
                            self.old_subtechnique_of_rels,
                        )
                    )
                )
                > 0
            )

    def get_groupings(
        self,
        obj_type,
        items,
        subtechnique_of_rels,
        id_to_technique,
        datacomponents,
        id_to_datasource,
    ):
        # get parents which have children
        if obj_type != "datasource":
            childless = list(
                filter(
                    lambda item: not self.has_subtechniques(item, True)
                    and not (
                        "x_mitre_is_subtechnique" in item
                        and item["x_mitre_is_subtechnique"]
                    ),
                    items,
                )
            )
            parents = list(
                filter(
                    lambda item: self.has_subtechniques(item, True)
                    and not (
                        "x_mitre_is_subtechnique" in item
                        and item["x_mitre_is_subtechnique"]
                    ),
                    items,
                )
            )
            children = {
                item["id"]: item
                for item in filter(
                    lambda item: ("x_mitre_is_subtechnique") in item
                    and (item["x_mitre_is_subtechnique"]),
                    items,
                )
            }
        else:
            childless = (
                []
            )  # all data sources should have data components, i.e., should have children
            parents = list(
                filter(
                    lambda item: not (
                        "x_mitre_data_source_ref" in item
                        and item["x_mitre_data_source_ref"]
                    ),
                    items,
                )
            )
            children = {
                item["id"]: item
                for item in filter(
                    lambda item: ("x_mitre_data_source_ref") in item
                    and (item["x_mitre_data_source_ref"]),
                    items,
                )
            }

        # stixID => [ children ]
        parentToChildren = {}
        for relationship in subtechnique_of_rels:
            if relationship["target_ref"] in parentToChildren:
                if relationship["source_ref"] in children:
                    parentToChildren[relationship["target_ref"]].append(
                        children[relationship["source_ref"]]
                    )
            else:
                if relationship["source_ref"] in children:
                    parentToChildren[relationship["target_ref"]] = [
                        children[relationship["source_ref"]]
                    ]

        for datacomponent in datacomponents:
            if datacomponent["x_mitre_data_source_ref"] in parentToChildren:
                if datacomponent["id"] in children:
                    parentToChildren[datacomponent["x_mitre_data_source_ref"]].append(
                        children[datacomponent["id"]]
                    )
            else:
                if datacomponent["id"] in children:
                    parentToChildren[datacomponent["x_mitre_data_source_ref"]] = [
                        children[datacomponent["id"]]
                    ]

        # now group parents and children
        groupings = []
        for parent in childless + parents:
            parent_children = (
                parentToChildren.pop(parent["id"])
                if parent["id"] in parentToChildren
                else []
            )
            groupings.append(
                {
                    "parent": parent,
                    "parentInSection": True,
                    "children": parent_children,
                }
            )

        for parentID in parentToChildren:
            if id_to_technique.get(parentID):
                parentObj = id_to_technique[parentID]
            elif id_to_datasource.get(parentID):
                parentObj = id_to_datasource[parentID]

            if parentObj:
                groupings.append(
                    {
                        "parent": parentObj,
                        "parentInSection": False,
                        "children": parentToChildren[parentID],
                    }
                )

        groupings = sorted(groupings, key=lambda grouping: grouping["parent"]["name"])
        return groupings

    def get_markdown_string(self):
        """Return a markdown string summarizing detected differences."""

        def getSectionList(items, obj_type, section):
            """Parse a list of items in a section and return a string for the items."""
            logger.debug(f"getting section list for {obj_type}/{section}")

            if section == "deletions":
                subtechnique_of_rels = self.old_subtechnique_of_rels
                id_to_technique = self.old_id_to_technique
                datacomponents = self.old_datacomponents
                id_to_datasource = self.old_id_to_datasource
            else:
                subtechnique_of_rels = self.new_subtechnique_of_rels
                id_to_technique = self.new_id_to_technique
                datacomponents = self.new_datacomponents
                id_to_datasource = self.new_id_to_datasource

            def placard(item):
                """Get a section list item for the given SDO according to section type"""
                if section == "revocations":
                    revoker = item["revoked_by"]
                    if (
                        "x_mitre_is_subtechnique" in revoker
                        and revoker["x_mitre_is_subtechnique"]
                    ):
                        # get revoking technique's parent for display
                        parentID = list(
                            filter(
                                lambda rel: rel["source_ref"] == revoker["id"],
                                subtechnique_of_rels,
                            )
                        )[0]["target_ref"]
                        parentName = (
                            id_to_technique[parentID]["name"]
                            if parentID in id_to_technique
                            else "ERROR NO PARENT"
                        )
                        return f"{item['name']} (revoked by { parentName}: [{revoker['name']}]({self.site_prefix}/{self.getUrlFromStix(revoker, True)}))"
                    elif (
                        "x_mitre_data_source_ref" in revoker
                        and revoker["x_mitre_data_source_ref"]
                    ):
                        # get revoking technique's parent for display
                        parentID = list(
                            filter(
                                lambda rel: rel["id"] == revoker["id"], datacomponents
                            )
                        )[0]["x_mitre_data_source_ref"]
                        parentName = (
                            id_to_datasource[parentID]["name"]
                            if parentID in id_to_datasource
                            else "ERROR NO PARENT"
                        )
                        return f"{item['name']} (revoked by { parentName}: [{revoker['name']}]({self.site_prefix}/{self.getDataComponentUrl(id_to_datasource[parentID], item)}))"
                    else:
                        return f"{item['name']} (revoked by [{revoker['name']}]({self.site_prefix}/{self.getUrlFromStix(revoker)}))"
                elif section == "deletions":
                    return f"{item['name']}"
                else:
                    is_subtechnique = (
                        item["type"] == "attack-pattern"
                        and "x_mitre_is_subtechnique" in item
                        and item["x_mitre_is_subtechnique"]
                    )
                    if item["type"] == "x-mitre-data-component":
                        parentID = item["x_mitre_data_source_ref"]
                        if id_to_datasource.get(parentID):
                            return f"[{item['name']}]({self.site_prefix}/{self.getDataComponentUrl(id_to_datasource[parentID], item)})"
                    return f"[{item['name']}]({self.site_prefix}/{self.getUrlFromStix(item, is_subtechnique)})"

            groupings = self.get_groupings(
                obj_type=obj_type,
                items=items,
                subtechnique_of_rels=subtechnique_of_rels,
                id_to_technique=id_to_technique,
                datacomponents=datacomponents,
                id_to_datasource=id_to_datasource,
            )

            # build sectionList string
            sectionString = ""
            for grouping in groupings:
                if grouping["parentInSection"]:
                    sectionString += f"* { placard(grouping['parent']) }\n"

                for child in sorted(
                    grouping["children"], key=lambda child: child["name"]
                ):
                    if grouping["parentInSection"]:
                        sectionString += f"  * {placard(child) }\n"
                    else:
                        sectionString += (
                            f"* {grouping['parent']['name']}: { placard(child) }\n"
                        )

            logger.debug(f"finished getting section list for {obj_type}/{section}")
            # logger.debug(sectionString)
            return sectionString

        def getContributorSection():
            # Get contributors markdown
            contribSection = "### Contributors to this release\n\n"
            sorted_contributors = sorted(
                self.release_contributors, key=lambda v: v.lower()
            )

            for contributor in sorted_contributors:
                if contributor == "ATT&CK":
                    continue  # do not include ATT&CK as contributor
                contribSection += f"* {contributor}\n"

            return contribSection

        logger.info("generating markdown string")
        content = ""
        for obj_type in self.data.keys():
            domains = ""
            for domain in self.data[obj_type]:
                logger.debug(
                    f"==== Generating markdown for domain: {domainToDomainLabel[domain]} --- {obj_type} ===="
                )
                domains += f"#### {domainToDomainLabel[domain]}\n\n"  # e.g "Enterprise"
                # Skip mobile sections for data sources
                if domain == "mobile-attack" and obj_type == "datasource":
                    logger.debug(
                        f"Skipping - ATT&CK for Mobile does not support data sources"
                    )
                    domains += "ATT&CK for Mobile does not support data sources\n\n"
                    continue
                domain_sections = ""
                for section, values in self.data[obj_type][domain].items():
                    logger.debug(f"{section}: {len(values)}")

                    if values:  # if there are items in the section
                        section_items = getSectionList(
                            items=values, obj_type=obj_type, section=section
                        )
                    else:  # no items in section
                        section_items = "* No changes\n"

                    header = sectionNameToSectionHeaders[section] + ":"

                    if "{obj_type}" in header:
                        if section == "additions":
                            header = header.replace(
                                "{obj_type}", attackTypeToTitle[obj_type]
                            )
                        else:
                            header = header.replace(
                                "{obj_type}", attackTypeToSectionName[obj_type]
                            )

                    # e.g "added techniques:"
                    domain_sections += f"{header}\n\n{section_items}\n"

                # add domain sections
                domains += f"{domain_sections}"

            # e.g "techniques"
            content += f"### {attackTypeToTitle[obj_type]}\n\n{domains}"

        if self.show_key:
            key_content = self.get_md_key()
            content = f"{key_content}\n\n{content}"

        # Add contributors if requested by argument
        if self.include_contributors:
            content += getContributorSection()

        logger.info("finished generating markdown string")

        return content

    def get_layers_dict(self):
        """Return ATT&CK Navigator layers in dict format summarizing detected differences.

        Returns a dict mapping domain to its layer dict.
        """
        logger.info("generating layers dict")

        layers = {}
        thedate = datetime.datetime.today().strftime("%B %Y")
        # for each layer file in the domains mapping
        for domain in self.domains:
            logger.debug(f"===== Generating layer for domain: {domain} =====")
            # build techniques list
            techniques = []
            used_statuses = set()
            for status in self.data["technique"][domain]:
                logger.debug(f"Parsing: {status}")
                if status == "revocations" or status == "deprecations":
                    continue
                for technique in self.data["technique"][domain][status]:
                    problem_detected = False
                    if "kill_chain_phases" not in technique:
                        logger.error(
                            f"{technique['id']}: technique missing a tactic!! {technique['name']}"
                        )
                        problem_detected = True
                    if "external_references" not in technique:
                        logger.error(
                            f"{technique['id']}: technique missing external references!! {technique['name']}"
                        )
                        problem_detected = True

                    if problem_detected:
                        continue

                    for phase in technique["kill_chain_phases"]:
                        techniques.append(
                            {
                                "techniqueID": technique["external_references"][0][
                                    "external_id"
                                ],
                                "tactic": phase["phase_name"],
                                "enabled": True,
                                "color": statusToColor[status],
                                "comment": status[:-1]
                                if status != "unchanged"
                                else status,  # trim s off end of word
                            }
                        )
                        used_statuses.add(status)

            # build legend based off used_statuses
            legendItems = list(
                map(
                    lambda status: {
                        "color": statusToColor[status],
                        "label": status + ": " + statusDescriptions[status],
                    },
                    used_statuses,
                )
            )

            # build layer structure
            layer_json = {
                "versions": {"layer": "4.1", "navigator": "4.1"},
                "name": f"{thedate} {domainToDomainLabel[domain]} Updates",
                "description": f"{domainToDomainLabel[domain]} updates for the {thedate} release of ATT&CK",
                "domain": domain,
                "techniques": techniques,
                "sorting": 0,
                "hideDisabled": False,
                "legendItems": legendItems,
                "showTacticRowBackground": True,
                "tacticRowBackground": "#205b8f",
                "selectTechniquesAcrossTactics": True,
            }
            layers[domain] = layer_json

        logger.info("finished generating layers dict")
        return layers

    def get_changes_dict(self):
        """Return dict format summarizing detected differences."""
        logger.info("generating changes dict")

        def cleanup_values(items, obj_type, section):
            if section == "deletions":
                subtechnique_of_rels = self.old_subtechnique_of_rels
                id_to_technique = self.old_id_to_technique
                datacomponents = self.old_datacomponents
                id_to_datasource = self.old_id_to_datasource
            else:
                subtechnique_of_rels = self.new_subtechnique_of_rels
                id_to_technique = self.new_id_to_technique
                datacomponents = self.new_datacomponents
                id_to_datasource = self.new_id_to_datasource

            groupings = self.get_groupings(
                obj_type=obj_type,
                items=items,
                subtechnique_of_rels=subtechnique_of_rels,
                id_to_technique=id_to_technique,
                datacomponents=datacomponents,
                id_to_datasource=id_to_datasource,
            )

            new_values = []
            for grouping in groupings:
                if grouping["parentInSection"]:
                    new_values.append(grouping["parent"])

                for child in sorted(
                    grouping["children"], key=lambda child: child["name"]
                ):
                    new_values.append(child)

            return new_values

        changes_dict = {}
        for domain in self.domains:
            changes_dict[domain] = {}

        for obj_type, domains in self.data.items():
            for domain, sections in domains.items():
                logger.debug(
                    f"===== Generating domain: {domainToDomainLabel[domain]} --- {obj_type} ====="
                )
                changes_dict[domain][obj_type] = {}

                for section, values in sections.items():
                    # new_values includes parents & children mixed (e.g. techniques/sub-techniques, data sources/components)
                    new_values = cleanup_values(
                        items=values, obj_type=obj_type, section=section
                    )
                    changes_dict[domain][obj_type][section] = new_values

        # always add contributors
        changes_dict["new-contributors"] = []
        sorted_contributors = sorted(self.release_contributors, key=lambda v: v.lower())
        for contributor in sorted_contributors:
            if contributor == "ATT&CK":
                continue  # do not include ATT&CK as contributor
            changes_dict["new-contributors"].append(contributor)

        logger.info("finished generating changes dict")
        return changes_dict


def markdown_to_index_html(markdown_outfile, content):
    """Convert the markdown string passed in to HTML and store in index.html
    of indicated output file path"""
    logger.info("writing HTML to file")

    # get output file path
    outputfile_path = os.path.split(markdown_outfile)[0]
    outfile = os.path.join(outputfile_path, "index.html")

    # Center content
    html_string = """<div style='max-width: 55em;margin: auto;margin-top:20px;font-family: "Roboto", sans-serif;'>"""
    html_string += "<meta charset='utf-8'>"
    html_string += (
        "<h1 style='text-align:center;'>Changes between ATT&CK STIX bundles</h1>"
    )
    html_string += markdown.markdown(content)
    html_string += "</div>"

    outfile = open(outfile, "w", encoding="utf-8")
    outfile.write(html_string)
    outfile.close()

    logger.info("finished writing HTML to file")


def layers_dict_to_files(outfiles, layers):
    """Print the layers dict passed in to layer files."""
    logger.info("writing layers dict to layer files")

    # write each layer to separate files
    if "enterprise-attack" in layers:
        enterprise_attack_layer_file = outfiles[0]
        Path(enterprise_attack_layer_file).parent.mkdir(parents=True, exist_ok=True)
        json.dump(layers["enterprise-attack"], open(enterprise_attack_layer_file, "w"), indent=4)

    if "mobile-attack" in layers:
        mobile_attack_layer_file = outfiles[1]
        Path(mobile_attack_layer_file).parent.mkdir(parents=True, exist_ok=True)
        json.dump(layers["mobile-attack"], open(mobile_attack_layer_file, "w"), indent=4)

    logger.info("finished writing layers dict to layer files")


def get_parsed_args():
    """Create argument parser and parse arguments"""
    old_dir_default = "old"

    parser = argparse.ArgumentParser(
        description=(
            "Create -markdown and/or -layers reporting on the changes between two versions of the ATT&CK content. "
            "Takes STIX bundles as input. For default operation, put enterprise-attack.json and mobile-attack.json bundles "
            "in 'old' and 'new' folders for the script to compare."
        )
    )

    parser.add_argument(
        "-old",
        type=str,
        metavar="OLD_DIR",
        help=f"the directory of the old content. Default is '{old_dir_default}'",
    )

    parser.add_argument(
        "-new",
        type=str,
        metavar="NEW_DIR",
        default="new",
        help="the directory of the new content. Default is '%(default)s'",
    )

    parser.add_argument(
        "-types",
        type=str,
        nargs="+",
        metavar=("OBJ_TYPE", "OBJ_TYPE"),
        choices=["technique", "software", "group", "mitigation", "datasource"],
        default=["technique", "software", "group", "mitigation", "datasource"],
        help="which types of objects to report on. Choices (and defaults) are %(choices)s",
    )

    parser.add_argument(
        "-domains",
        type=str,
        nargs="+",
        metavar="DOMAIN",
        choices=["enterprise-attack", "mobile-attack"],
        default=["enterprise-attack", "mobile-attack"],
        help="which domains to report on. Choices (and defaults) are %(choices)s",
    )

    parser.add_argument(
        "-markdown",
        type=str,
        nargs="?",
        metavar="MARKDOWN_FILE",
        const=md_default,
        help="create a markdown file reporting changes. If value is unspecified, defaults to %(const)s",
    )

    parser.add_argument(
        "-json-output",
        type=str,
        nargs="?",
        metavar="JSON_FILE",
        const=json_default,
        help="create a JSON file reporting changes. If value is unspecified, defaults to %(const)s",
    )

    parser.add_argument(
        "--create-html",
        action="store_true",
        help="create index.html page of markdown file that reported changes. Does not do anything unless -markdown is provided",
    )

    parser.add_argument(
        "-layers",
        type=str,
        nargs="*",
        # metavar=("ENTERPRISE", "MOBILE", "PRE"),
        help=f"""
            create layer files showing changes in each domain
            expected order of filenames is 'enterprise', 'mobile', 'pre attack'. 
            If values are unspecified, defaults to {", ".join(layer_defaults)}
            """,
    )

    parser.add_argument(
        "-site_prefix",
        type=str,
        default="",
        help="prefix links in markdown output, e.g. [prefix]/techniques/T1484",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print status messages",
    )

    parser.add_argument(
        "--minor-changes",
        action="store_true",
        help="show changes to objects which didn't increment the version number",
    )

    parser.add_argument(
        "--unchanged",
        action="store_true",
        help="show objects without changes in the markdown output",
    )

    parser.add_argument(
        "--use-taxii",
        action="store_true",
        help="Use content from the ATT&CK TAXII server for the -old data",
    )

    parser.add_argument(
        "--use-mitre-cti",
        action="store_true",
        help="Use content from the MITRE CTI repo for the -old data",
    )

    parser.add_argument(
        "--show-key",
        action="store_true",
        help="Add a key explaining the change types to the markdown",
    )

    parser.add_argument(
        "--contributors",
        action="store_true",
        help="show new contributors between releases",
    )

    args = parser.parse_args()

    # the default loguru logger logs up to Debug by default
    logger.remove()
    if args.verbose:
        logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)
    else:
        logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True, level="INFO")

    if args.use_taxii and args.old is not None:
        parser.error("--use-taxii and -old cannot be used together")

    if args.use_mitre_cti and args.old is not None:
        parser.error("--use-mitre-cti and -old cannot be used together")

    if not args.markdown and args.layers is None:
        logger.error(
            "Script doesn't output anything unless -markdown and/or -layers are specified."
        )
        logger.error("Run 'python3 diff_stix.py -h' for usage instructions")
        exit()

    if args.old is None:
        args.old = old_dir_default

    if args.layers is not None:
        if len(args.layers) not in [0, 3]:
            parser.error(
                "-layers requires exactly three files to be specified or none at all"
            )

    return args


# Used by attack-website script to generate changelog
def get_new_changelog_md(
    domains: List[str] = ["enterprise-attack", "mobile-attack"],
    layers: List[str] = layer_defaults,
    markdown_file: str = md_default,
    minor_changes: bool = False,
    unchanged: bool = False,
    new: str = "new",
    old: str = None,
    show_key: bool = False,
    site_prefix: str = "",
    types: List[str] = ["technique", "software", "group", "mitigation", "datasource"],
    use_taxii: bool = False,
    use_mitre_cti: bool = False,
    verbose: bool = False,
    include_contributors: bool = False,
    create_html: bool = False,
    json_output: str = json_default,
):
    # the default loguru logger logs up to Debug by default
    logger.remove()
    if verbose:
        logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)
    else:
        logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True, level="INFO")

    # if old:
    #     if use_mitre_cti or use_taxii:
    #         logger.error("Multiple sources selected as base STIX to compare against.")
    #         logger.error("When calling get_new_changelog_md(), 'old' is mutually exclusive with 'use_taxii' and 'use_mitre_cti'")
    #         return ""

    diffStix = DiffStix(
        domains=domains,
        layers=layers,
        markdown=markdown_file,
        minor_changes=minor_changes,
        unchanged=unchanged,
        new=new,
        old=old,
        show_key=show_key,
        site_prefix=site_prefix,
        types=types,
        use_taxii=use_taxii,
        use_mitre_cti=use_mitre_cti,
        verbose=verbose,
        include_contributors=include_contributors,
    )

    md_string = None
    if markdown_file:
        md_string = diffStix.get_markdown_string()

        logger.info("writing markdown to file")
        Path(markdown_file).parent.mkdir(parents=True, exist_ok=True)
        with open(markdown_file, "w") as file:
            file.write(md_string)
        logger.info("finished writing markdown to file")

        if create_html:
            markdown_to_index_html(markdown_file, md_string)

    if layers:
        if len(layers) == 0:
            # no files specified, e.g. '-layers', use defaults
            diffStix.layers = layer_defaults
            layers = layer_defaults
        elif len(layers) == 3:
            # files specified, e.g. '-layers file.json file2.json file3.json', use specified
            # assumes order of files is enterprise, mobile, pre attack (same order as defaults)
            diffStix.layers = layers

        layers_dict = diffStix.get_layers_dict()
        layers_dict_to_files(outfiles=layers, layers=layers_dict)

    if json_output:
        changes_dict = diffStix.get_changes_dict()

        logger.info("writing JSON updates to file")
        Path(json_output).parent.mkdir(parents=True, exist_ok=True)
        json.dump(changes_dict, open(json_output, "w"), indent=4)
        logger.info("finished writing JSON updates to file")

    return md_string


def main():
    args = get_parsed_args()

    get_new_changelog_md(
        domains=args.domains,
        layers=args.layers,
        markdown_file=args.markdown,
        minor_changes=args.minor_changes,
        unchanged=args.unchanged,
        new=args.new,
        old=args.old,
        show_key=args.show_key,
        site_prefix=args.site_prefix,
        types=args.types,
        use_taxii=args.use_taxii,
        use_mitre_cti=args.use_mitre_cti,
        verbose=args.verbose,
        include_contributors=args.contributors,
        create_html=args.create_html,
        json_output=args.json_output,
    )


if __name__ == "__main__":
    main()



import json
from copy import deepcopy

from mitreattack.navlayers.core import Layer
from mitreattack.navlayers.exporters.svg_templates import SvgTemplates


class NoLayer(Exception):
    pass


class SVGConfig:
    d_width = 8.5
    d_height = 11
    d_headerHeight = 1
    d_unit = "in"
    d_showSubtechniques = "expanded"
    d_font = "sans-serif"
    d_tableBorderColor = "#6B7279"
    d_showHeader = True
    d_legendDocked = True
    d_legendX = 0
    d_legendY = 0
    d_legendWidth = 2
    d_legendHeight = 1
    d_showLegend = True
    d_showFilters = True
    d_showAbout = True
    d_showDomain = True
    d_border = 0.104

    def __init__(self, width=d_width, height=d_height, headerHeight=d_headerHeight, unit=d_unit,
                 showSubtechniques=d_showSubtechniques, font=d_font, tableBorderColor=d_tableBorderColor,
                 showHeader=d_showHeader, legendDocked=d_legendDocked, legendX=d_legendX, legendY=d_legendY,
                 legendWidth=d_legendWidth, legendHeight=d_legendHeight, showLegend=d_showLegend,
                 showFilters=d_showFilters, showAbout=d_showAbout, showDomain=d_showDomain, border=d_border):
        """
            Define parameters to configure SVG export

            :param width: Desired SVG width
            :param height: Desired SVG height
            :param headerHeight: Desired Header Block height
            :param unit: SVG measurement units (qualifies width, height, etc.) - "in", "cm", "px", "em", or "pt"
            :param showSubtechniques: Display form for subtechniques - "all", "expanded" (decided by layer), or "none"
            :param font: What font style to use - "serif", "sans-serif", or "monospace"
            :param tableBorderColor: Hex color to use for the technique borders
            :param showHeader: Whether or not to show Header Blocks
            :param legendDocked: Whether or not the legend should be docked
            :param legendX: Where to place the legend on the x axis if not docked
            :param legendY: Where to place the legend on the y axis if not docked
            :param legendWidth: Width of the legend if not docked
            :param legendHeight: Height of the legend if not docked
            :param showLegend: Whether or not to show the legend
            :param showFilters: Whether or not to show the Filter Header Block
            :param showAbout: Whether or not to show the About Header Block
            :param showDomain: Whether or not to show the Domain Version Header Block
            :param border: What default border width to use
        """
        # force defaults in case bad values are provided so we don't crash later
        self.width = self.d_width
        self.height = self.d_height
        self.headerHeight = self.d_headerHeight
        self.unit = self.d_unit
        self.showSubtechniques = self.d_showSubtechniques
        self.font = self.d_font
        self.tableBorderColor = self.d_tableBorderColor
        self.showHeader = self.d_showHeader
        self.legendDocked = self.d_legendDocked
        self.legendX = self.d_legendX
        self.legendY = self.d_legendY
        self.legendWidth = self.d_legendWidth
        self.legendHeight = self.d_legendHeight
        self.showDomain = self.d_showDomain
        self.showLegend = self.d_showLegend
        self.showFilters = self.d_showFilters
        self.showAbout = self.d_showAbout
        self.border = self.d_border

        self.width = width
        self.height = height
        self.headerHeight = headerHeight
        self.unit = unit
        self.showSubtechniques = showSubtechniques
        self.font = font
        self.tableBorderColor = tableBorderColor
        self.showHeader = showHeader
        self.legendDocked = legendDocked
        self.legendX = legendX
        self.legendY = legendY
        self.legendWidth = legendWidth
        self.legendHeight = legendHeight
        self.showDomain = showDomain
        self.showLegend = showLegend
        self.showFilters = showFilters
        self.showAbout = showAbout
        self.border = border

    def load_from_file(self, filename=""):
        """
            Load config from a json file

            :param filename: The file to read from
        """
        with open(filename, 'r') as fio:
            raw = fio.read()
        self._data = json.loads(raw)
        for entry in self._data:
            patched = entry
            if not patched.startswith('_SVGConfig__'):
                patched = '_SVGConfig__' + patched
            if patched in vars(self).keys():
                setattr(self, entry, self._data[entry])
            else:
                print('WARNING - Unidentified Config Field in {}: {}'.format(filename, entry))

        self.__str__()

    def save_to_file(self, filename=""):
        """
            Store config to json file

            :param filename: The file to write to
        """
        out = dict(width=self.width, height=self.height, headerHeight=self.headerHeight, unit=self.unit,
                   showSubtechniques=self.showSubtechniques, font=self.font, tableBorderColor=self.tableBorderColor,
                   showHeader=self.showHeader, legendDocked=self.legendDocked, legendX=self.legendX,
                   legendY=self.legendY, legendWidth=self.legendWidth, legendHeight=self.legendHeight,
                   showLegend=self.showLegend, showFilters=self.showFilters, showAbout=self.showAbout,
                   border=self.border)
        with open(filename, 'w', encoding="utf-16") as file:
            json.dump(out, file, ensure_ascii=False)

    def __str__(self):
        """
            INTERNAL - display current configuration
        """
        print("SVGConfig current settings: ")
        print("width - {}".format(self.width))
        print("height - {}".format(self.height))
        print("headerHeight - {}".format(self.headerHeight))
        print("unit - {}".format(self.unit))
        print("showSubtechniques - {}".format(self.showSubtechniques))
        print("font - {}".format(self.font))
        print("tableBorderColor - {}".format(self.tableBorderColor))
        print("showHeader - {}".format(self.showHeader))
        print("legendDocked - {}".format(self.legendDocked))
        print("legendX - {}".format(self.legendX))
        print("legendY - {}".format(self.legendY))
        print("legendWidth - {}".format(self.legendWidth))
        print("legendHeight - {}".format(self.legendHeight))
        print("showLegend - {}".format(self.showLegend))
        print("showFilters - {}".format(self.showFilters))
        print("showAbout - {}".format(self.showAbout))
        print("border - {}".format(self.border))

    @property
    def width(self):
        if self.__width is not None:
            return self.__width

    @width.setter
    def width(self, width):
        if isinstance(width, int) or isinstance(width, float):
            self.__width = width
        else:
            print('[Warning] - Unable to set width to {}: not a float or int'.format(width))

    @property
    def height(self):
        if self.__height is not None:
            return self.__height

    @height.setter
    def height(self, height):
        if isinstance(height, int) or isinstance(height, float):
            self.__height = height
        else:
            print('[Warning] - Unable to set height to {}: not a float or int'.format(height))

    @property
    def headerHeight(self):
        if self.__headerHeight is not None:
            return self.__headerHeight

    @headerHeight.setter
    def headerHeight(self, headerHeight):
        if isinstance(headerHeight, int) or isinstance(headerHeight, float):
            self.__headerHeight = headerHeight
        else:
            print('[Warning] - Unable to set headerHeight to {}: not a float or int'.format(headerHeight))

    @property
    def unit(self):
        if self.__unit is not None:
            return self.__unit

    @unit.setter
    def unit(self, unit):
        if unit in ['in', 'cm', 'px', 'em', 'pt']:
            self.__unit = unit
        else:
            print('[Warning] - Unable to set unit to {}: not one of ["in", "cm", "px", "em", "pt"]'.format(unit))

    @property
    def showSubtechniques(self):
        if self.__showSubtechniques is not None:
            return self.__showSubtechniques

    @showSubtechniques.setter
    def showSubtechniques(self, showSubtechniques):
        if showSubtechniques in ["expanded", "all", "none"]:
            self.__showSubtechniques = showSubtechniques
        else:
            print('[Warning] - Unable to set showSubtechniques to {}: not one of ["expanded", "all", "none"]'.format(
                showSubtechniques))

    @property
    def font(self):
        if self.__font is not None:
            return self.__font

    @font.setter
    def font(self, font):
        if font in ["serif", "sans-serif", "monospace"]:
            self.__font = font
        else:
            print('[Warning] - Unable to set font to {}: not one of ["serif", "sans-serif", "monospace"]'.format(font))

    @property
    def tableBorderColor(self):
        if self.__tableBorderColor is not None:
            return self.__tableBorderColor

    @tableBorderColor.setter
    def tableBorderColor(self, tableBorderColor):
        if isinstance(tableBorderColor, str) and tableBorderColor.startswith('#') and len(tableBorderColor) == 7:
            self.__tableBorderColor = tableBorderColor
        else:
            reason = ''
            if not isinstance(tableBorderColor, str):
                reason = 'not a string'
            elif not tableBorderColor.startswith('#'):
                reason = 'not a valid code (does not start with #)'
            elif len(tableBorderColor) != 7:
                reason = 'not a valid code (#ZZZZZZ)'
            print('[Warning] - Unable to set tableBorderColor to {}: '.format(tableBorderColor) + reason)

    @property
    def showHeader(self):
        if self.__showHeader is not None:
            return self.__showHeader

    @showHeader.setter
    def showHeader(self, showHeader):
        if isinstance(showHeader, bool):
            self.__showHeader = showHeader
        else:
            print('[Warning] - Unable to set showHeader to {}: not a bool'.format(showHeader))

    @property
    def legendDocked(self):
        if self.__legendDocked is not None:
            return self.__legendDocked

    @legendDocked.setter
    def legendDocked(self, legendDocked):
        if isinstance(legendDocked, bool):
            self.__legendDocked = legendDocked
        else:
            print('[Warning] - Unable to set legendDocked to {}: not a bool'.format(legendDocked))

    @property
    def legendX(self):
        if self.__legendX is not None:
            return self.__legendX

    @legendX.setter
    def legendX(self, legendX):
        if isinstance(legendX, int) or isinstance(legendX, float):
            self.__legendX = legendX
        else:
            print('[Warning] - Unable to set legendX to {}: not a float or int'.format(legendX))

    @property
    def legendY(self):
        if self.__legendY is not None:
            return self.__legendY

    @legendY.setter
    def legendY(self, legendY):
        if isinstance(legendY, int) or isinstance(legendY, float):
            self.__legendY = legendY
        else:
            print('[Warning] - Unable to set legendY to {}: not a float or int'.format(legendY))

    @property
    def legendWidth(self):
        if self.__legendWidth is not None:
            return self.__legendWidth

    @legendWidth.setter
    def legendWidth(self, legendWidth):
        if isinstance(legendWidth, int) or isinstance(legendWidth, float):
            self.__legendWidth = legendWidth
        else:
            print('[Warning] - Unable to set legendWidth to {}: not a float or int'.format(legendWidth))

    @property
    def legendHeight(self):
        if self.__legendHeight is not None:
            return self.__legendHeight

    @legendHeight.setter
    def legendHeight(self, legendHeight):
        if isinstance(legendHeight, int) or isinstance(legendHeight, float):
            self.__legendHeight = legendHeight
        else:
            print('[Warning] - Unable to set legendHeight to {}: not a float or int'.format(legendHeight))

    @property
    def showLegend(self):
        if self.__showLegend is not None:
            return self.__showLegend

    @showLegend.setter
    def showLegend(self, showLegend):
        if isinstance(showLegend, bool):
            self.__showLegend = showLegend
        else:
            print('[Warning] - Unable to set showLegend to {}: not a bool'.format(showLegend
                                                                                  ))

    @property
    def showFilters(self):
        if self.__showFilters is not None:
            return self.__showFilters

    @showFilters.setter
    def showFilters(self, showFilters):
        if isinstance(showFilters, bool):
            self.__showFilters = showFilters
        else:
            print('[Warning] - Unable to set showFilters to {}: not a bool'.format(showFilters))

    @property
    def showAbout(self):
        if self.__showAbout is not None:
            return self.__showAbout

    @showAbout.setter
    def showAbout(self, showAbout):
        if isinstance(showAbout, bool):
            self.__showAbout = showAbout
        else:
            print('[Warning] - Unable to set showAbout to {}: not a bool'.format(showAbout))

    @property
    def showDomain(self):
        if self.__showDomain is not None:
            return self.__showDomain

    @showDomain.setter
    def showDomain(self, showDomain):
        if isinstance(showDomain, bool):
            self.__showDomain = showDomain
        else:
            print('[Warning] - Unable to set showAbout to {}: not a bool'.format(showDomain))

    @property
    def border(self):
        if self.__border is not None:
            return self.__border

    @border.setter
    def border(self, border):
        if isinstance(border, float):
            self.__border = border
        else:
            print('[Warning] - Unable to set border to {}: not a float'.format(border))


class ToSvg:
    def __init__(self, domain='enterprise', source='taxii', resource=None, config=None):
        """
            Sets up exporting system, builds underlying matrix

            :param domain: Which domain to utilize for the underlying matrix layout
            :param source: Use the taxii server, local data, or a remote ATT&CK Workbench instance
            :param resource: string path to local cache of stix data (local) or url of workbench to reach out 
                                to (remote)
            :param config: Optional pre-existing SVGConfig object
        """
        self.raw_handle = SvgTemplates(domain=domain, source=source, resource=resource)
        if config is not None and isinstance(config, SVGConfig):
            self.config = config
        else:
            self.config = SVGConfig()

    def to_svg(self, layerInit, filepath='example.svg'):
        """
            Generate a svg file based on the input layer

            :param layerInit: Input attack layer object to transform
            :param filepath: Desired output svg location
            :return: (meta) svg file at the targeted output location
        """
        if layerInit is not None:
            if not isinstance(layerInit, Layer):
                raise TypeError

        if layerInit is None:
            raise NoLayer

        layer = deepcopy(layerInit)

        included_subs = []
        if layer.layer.techniques:
            for entry in layer.layer.techniques:
                if self.config.showSubtechniques == "expanded":
                    if entry.showSubtechniques:
                        if entry.tactic:
                            included_subs.append((entry.techniqueID, entry.tactic))
                        else:
                            included_subs.append((entry.techniqueID, False))
                elif self.config.showSubtechniques == "all":
                    if entry.tactic:
                        included_subs.append((entry.techniqueID, entry.tactic))
                    else:
                        included_subs.append((entry.techniqueID, False))
                else:  # none displayed
                    pass

        excluded = []
        if layer.layer.hideDisabled:
            for entry in layer.layer.techniques:
                if entry.enabled is False:
                    if entry.tactic:
                        excluded.append((entry.techniqueID, entry.tactic))
                    else:
                        excluded.append((entry.techniqueID, False))
        scores = []
        colors = []
        if layer.layer.techniques:
            for entry in layer.layer.techniques:
                tscore = entry.score
                if tscore is not None:
                    if entry.tactic:
                        scores.append((entry.techniqueID, entry.tactic, tscore))
                    else:
                        scores.append((entry.techniqueID, False, tscore))
                elif entry.color:
                    if entry.tactic:
                        colors.append((entry.techniqueID, entry.tactic, entry.color))
                    else:
                        colors.append((entry.techniqueID, False, entry.color))
        sName = True
        sID = False
        sort = 0
        if layer.layer.layout:
            sName = layer.layer.layout.showName
            sID = layer.layer.layout.showID
        if layer.layer.sorting:
            sort = layer.layer.sorting
        d = self.raw_handle.export(showName=sName, showID=sID, sort=sort, scores=scores,
                                   subtechs=included_subs, colors=colors,
                                   exclude=excluded, lhandle=layer.layer, config=self.config)
        d.saveSvg(filepath)



import argparse

from mitreattack.navlayers.exporters.to_svg import ToSvg, SVGConfig
from mitreattack.navlayers.exporters.to_excel import ToExcel
from mitreattack.navlayers.core import Layer


def main(argv=None):

    parser = argparse.ArgumentParser(description='Export an ATT&CK Navigator layer as a svg image or excel file')
    parser.add_argument('-m', '--mode', choices=['svg', 'excel'], required=True,
                        help='The form to export the layers in')
    parser.add_argument('input', nargs='+', help='Path(s) to the file to export')
    parser.add_argument('-s', '--source', choices=['taxii', 'local', 'remote'], default='taxii',
                        help='What source to utilize when building the matrix')
    parser.add_argument('--resource', help='Path to the local resource if --source=local, or url of an ATT&CK Workbench'
                                           ' instance if --source=remote', default=None)
    parser.add_argument('-o', '--output', nargs='+', help='Path(s) to the exported svg/xlsx file', required=True)
    parser.add_argument('-l', '--load_settings', help='[SVG Only] Path to a SVG configuration json to use when '
                                                      'rendering', default=None)
    parser.add_argument('-d', '--size', nargs=2, help='[SVG Only] X and Y size values (in inches) for SVG export (use '
                                                      '-l for other settings)', default=[8.5, 11], metavar=("WIDTH",
                                                                                                            "HEIGHT"))
    args = parser.parse_args(args=argv)

    if len(args.output) != len(args.input):
        print('Mismatched number of output file paths to input file paths. Exiting...')
        return

    for i in range(0, len(args.input)):
        entry = args.input[i]
        print('{}/{} - Beginning processing {}'.format(i + 1, len(args.input), entry))
        lay = Layer()
        try:
            lay.from_file(entry)
        except Exception as e:
            print('Unable to load {} due to exception: {}. Skipping...'.format(entry, e))
            continue
        if args.mode == 'excel':
            if not args.output[i].endswith('.xlsx'):
                print('[ERROR] Unable to export {} as type: excel to {}'.format(entry, args.output[i]))
                continue
            exy = ToExcel(domain=lay.layer.domain, source=args.source, resource=args.resource)
            exy.to_xlsx(lay, filepath=args.output[i])
        else:
            if not args.output[i].endswith('.svg'):
                print('[ERROR] Unable to export {} as type: svg to {}'.format(entry, args.output[i]))
                continue
            conf = SVGConfig()
            if args.load_settings:
                conf.load_from_file(args.load_settings)
            if len(args.size) == 2:
                conf.width = float(args.size[0])
                conf.height = float(args.size[1])
            svy = ToSvg(domain=lay.layer.domain, source=args.source, resource=args.resource, config=conf)
            svy.to_svg(lay, filepath=args.output[i])
        print('{}/{} - Finished exporting {} to {}'.format(i + 1, len(args.input), entry, args.output[i]))


if __name__ == '__main__':
    main()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mitreattack-python",
    version="1.5.1",
    author="MITRE ATT&CK, MITRE Corporation",
    author_email="attack@mitre.org",
    description="MITRE ATT&CK python library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer="Caleb Little",
    maintainer_email="clittle@mitre.org",
    license="Apache 2.0",
    url="https://github.com/mitre-attack/mitreattack-python/",
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'attackToExcel_cli=mitreattack.attackToExcel.attackToExcel:main',
            'layerExporter_cli=mitreattack.navlayers.layerExporter_cli:main',
            'layerGenerator_cli=mitreattack.navlayers.layerGenerator_cli:main',
            'indexToMarkdown_cli=mitreattack.collections.index_to_markdown:main',
            'collectionToIndex_cli=mitreattack.collections.collection_to_index:main',
            'stixToCollection_cli=mitreattack.collections.stix_to_collection:main',
            'diff_stix=mitreattack.diffStix.changelog_helper:main'
        ]
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Security",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'colour',
        'drawSvg',
        'loguru',
        'Markdown',
        'numpy',
        'openpyxl',
        'pandas',
        'Pillow',
        'requests',
        'stix2',
        'stix2-elevator',
        'tabulate',
        'taxii2-client',
        'tqdm',
        'xlsxwriter',
    ]
)



from stix2 import Filter, datastore
from mitreattack.navlayers.generators.usage_generator import UsageLayerGenerator
from mitreattack.navlayers.core.exceptions import typeChecker, categoryChecker
from mitreattack.navlayers.generators.gen_helpers import remove_revoked_depreciated, get_attack_id
from tqdm import tqdm


class BatchGenerator:
    def __init__(self, source, domain='enterprise', resource=None):
        """
        Initialize the Generator
        :param source: Which source to use for data (local, taxii [server], or [remote] ATT&CK Workbench)
        :param domain: Which matrix to use during generation
        :param resource: string path to local STIX data (local) or url of workbench to reach out to (remote)
        """
        self.usage_handle = UsageLayerGenerator(source, domain, resource)
        self.mapping = dict(group=[Filter('type', '=', 'intrusion-set')],
                            software=[Filter('type', '=', 'malware'), Filter('type', '=', 'tool')],
                            mitigation=[Filter('type', '=', 'course-of-action')],
                            datasource=[Filter('type', '=', 'x-mitre-data-source')])

    def generate_layers(self, layers_type):
        """
        Generate and return a collection of layers for all objects of a given type
        :param layers_type: the type of object to generate layers for (group, software, mitigation or datasource)
        :return: dictionary of generated layer objects, referenced by STIX-ID
        """
        typeChecker(type(self).__name__, layers_type, str, "type")
        categoryChecker(type(self).__name__, layers_type, ["group", "software", "mitigation", "datasource"], "type")
        produced = dict()
        object_listing = remove_revoked_depreciated(self.usage_handle.source_handle.query(self.mapping[layers_type]))
        for entry in tqdm(object_listing, desc=f"building {layers_type} matrices"):
            try:
                produced[entry['id']] = self.usage_handle.generate_layer(get_attack_id(entry))
            except datastore.DataSourceError as e:
                print(f"WARNING - unable to generate layer for {(entry['id'], entry['name'])}. "
                      f"Specifically, generator encountered {e}. Continuing...")
        return produced



from mitreattack.navlayers.core.filter import Filter
from mitreattack.navlayers.core.layout import Layout
from mitreattack.navlayers.core.technique import Technique
from mitreattack.navlayers.core.gradient import Gradient
from mitreattack.navlayers.core.legenditem import LegendItem
from mitreattack.navlayers.core.metadata import Metadata, MetaDiv
from mitreattack.navlayers.core.objlink import Link, LinkDiv
from mitreattack.navlayers.core.versions import Versions
from mitreattack.navlayers.core.exceptions import UNSETVALUE, typeChecker, handler, categoryChecker, \
     UnknownLayerProperty, loadChecker, MissingParameters
from mitreattack.navlayers.core.helpers import handle_object_placement


class _LayerObj:
    def __init__(self, name, domain):
        """
            Initialization - Creates a layer object

            :param name: The name for this layer
            :param domain: The domain for this layer (enterprise-attack
                or mobile-attack)
        """
        self.__versions = UNSETVALUE
        self.name = name
        self.__description = UNSETVALUE
        self.domain = domain
        self.__filters = UNSETVALUE
        self.__sorting = UNSETVALUE
        self.__layout = UNSETVALUE
        self.__hideDisabled = UNSETVALUE
        self.__techniques = UNSETVALUE
        self.__gradient = UNSETVALUE
        self.__legendItems = UNSETVALUE
        self.__showTacticRowBackground = UNSETVALUE
        self.__tacticRowBackground = UNSETVALUE
        self.__selectTechniquesAcrossTactics = UNSETVALUE
        self.__selectSubtechniquesWithParent = UNSETVALUE
        self.__metadata = UNSETVALUE
        self.__links = UNSETVALUE

    @property
    def version(self):
        if self.__versions != UNSETVALUE:
            return self.__versions.layer

    @version.setter
    def version(self, version):
        typeChecker(type(self).__name__, version, str, "version")
        categoryChecker(type(self).__name__, version, ["3.0", "4.0", "4.1", "4.2", "4.3"], "version")
        if self.__versions is UNSETVALUE:
            self.__versions = Versions()
        self.__versions.layer = version

    @property
    def versions(self):
        if self.__versions != UNSETVALUE:
            return self.__versions

    @versions.setter
    def versions(self, versions):
        ret = handle_object_placement(self.__versions, versions, Versions)
        if ret:
            self.__versions = ret
        else:
            typeChecker(type(self).__name__, versions, dict, "version")
            attack = UNSETVALUE
            if 'attack' in versions:
                attack = versions['attack']
            try:
                loadChecker(type(self).__name__, versions, ['layer', 'navigator'], "versions")
                self.__versions = Versions(versions['layer'], attack, versions['navigator'])
            except MissingParameters as e:
                handler(type(self).__name__, 'versions {} is missing parameters: '
                                             '{}. Skipping.'
                        .format(versions, e))

    @property
    def name(self):
        return self.__name

    @name.setter
    def name(self, name):
        typeChecker(type(self).__name__, name, str, "name")
        self.__name = name

    @property
    def domain(self):
        return self.__domain

    @domain.setter
    def domain(self, domain):
        typeChecker(type(self).__name__, domain, str, "domain")
        dom = domain
        if dom.startswith('mitre'):
            dom = dom.split('-')[-1] + '-attack'
        categoryChecker(type(self).__name__, dom, ["enterprise-attack",
                                                   "mobile-attack",
                                                   "ics-attack"],
                        "domain")
        self.__domain = domain

    @property
    def description(self):
        if self.__description != UNSETVALUE:
            return self.__description

    @description.setter
    def description(self, description):
        typeChecker(type(self).__name__, description, str, "description")
        self.__description = description

    @property
    def filters(self):
        if self.__filters != UNSETVALUE:
            return self.__filters

    @filters.setter
    def filters(self, filters):
        ret = handle_object_placement(self.__filters, filters, Filter)
        if ret:
            self.__filters = ret
        else:
            temp = Filter(self.domain)
            try:
                loadChecker(type(self).__name__, filters, ['platforms'], "filters")
                # force upgrade to v4
                if 'stages' in filters:
                    print('[Filters] - V3 Field "stages" detected. Upgrading Filters object to V4.')
                temp.platforms = filters['platforms']
                self.__filters = temp
            except MissingParameters as e:
                handler(type(self).__name__, 'Filters {} is missing parameters: '
                                             '{}. Skipping.'
                        .format(filters, e))

    @property
    def sorting(self):
        if self.__sorting != UNSETVALUE:
            return self.__sorting

    @sorting.setter
    def sorting(self, sorting):
        typeChecker(type(self).__name__, sorting, int, "sorting")
        categoryChecker(type(self).__name__, sorting, [0, 1, 2, 3], "sorting")
        self.__sorting = sorting

    @property
    def layout(self):
        if self.__layout != UNSETVALUE:
            return self.__layout

    @layout.setter
    def layout(self, layout):
        ret = handle_object_placement(self.__layout, layout, Layout)
        if ret:
            self.__layout = ret
        else:
            temp = Layout()
            if "layout" in layout:
                temp.layout = layout['layout']
            if "showName" in layout:
                temp.showName = layout['showName']
            if "showID" in layout:
                temp.showID = layout['showID']
            if "showAggregateScores" in layout:
                temp.showAggregateScores = layout['showAggregateScores']
            if "countUnscored" in layout:
                temp.countUnscored = layout['countUnscored']
            if 'aggregateFunction' in layout:
                temp.aggregateFunction = layout['aggregateFunction']
            self.__layout = temp

    @property
    def hideDisabled(self):
        if self.__hideDisabled != UNSETVALUE:
            return self.__hideDisabled

    @hideDisabled.setter
    def hideDisabled(self, hideDisabled):
        typeChecker(type(self).__name__, hideDisabled, bool, "hideDisabled")
        self.__hideDisabled = hideDisabled

    @property
    def techniques(self):
        if self.__techniques != UNSETVALUE:
            return self.__techniques

    @techniques.setter
    def techniques(self, techniques):
        typeChecker(type(self).__name__, techniques, list, "techniques")
        self.__techniques = []

        for entry in techniques:
            ret = handle_object_placement(self.__techniques, entry, Technique, list=True)
            if ret:
                self.__techniques = ret
            else:
                try:
                    loadChecker(type(self).__name__, entry, ['techniqueID'], "technique")
                    temp = Technique(entry['techniqueID'])
                    temp._loader(entry)
                    self.__techniques.append(temp)
                except MissingParameters as e:
                    handler(type(self).__name__, 'Technique {} is missing parameters: '
                                                 '{}. Skipping.'
                            .format(entry, e))

    @property
    def gradient(self):
        if self.__gradient != UNSETVALUE:
            return self.__gradient

    @gradient.setter
    def gradient(self, gradient):
        ret = handle_object_placement(self.__gradient, gradient, Gradient)
        if ret:
            self.__gradient = ret
        else:
            try:
                loadChecker(type(self).__name__, gradient, ['colors', 'minValue', 'maxValue'], "gradient")
                self.__gradient = Gradient(gradient['colors'], gradient['minValue'], gradient['maxValue'])
            except MissingParameters as e:
                handler(type(self).__name__, 'Gradient {} is missing parameters: '
                                             '{}. Skipping.'
                        .format(gradient, e))

    @property
    def legendItems(self):
        if self.__legendItems != UNSETVALUE:
            return self.__legendItems

    @legendItems.setter
    def legendItems(self, legendItems):
        typeChecker(type(self).__name__, legendItems, list, "legendItems")
        self.__legendItems = []
        for entry in legendItems:
            ret = handle_object_placement(self.__legendItems, entry, LegendItem, list=True)
            if ret:
                self.__legendItems = ret
            else:
                try:
                    loadChecker(type(self).__name__, entry, ['label', 'color'], "legendItem")
                    temp = LegendItem(entry['label'], entry['color'])
                    self.__legendItems.append(temp)
                except MissingParameters as e:
                    handler(type(self).__name__, 'Legend Item {} is missing parameters: '
                                                 '{}. Skipping.'
                            .format(entry, e))

    @property
    def showTacticRowBackground(self):
        if self.__showTacticRowBackground != UNSETVALUE:
            return self.__showTacticRowBackground

    @showTacticRowBackground.setter
    def showTacticRowBackground(self, showTacticRowBackground):
        typeChecker(type(self).__name__, showTacticRowBackground, bool,
                    "showTacticRowBackground")
        self.__showTacticRowBackground = showTacticRowBackground

    @property
    def tacticRowBackground(self):
        if self.__tacticRowBackground != UNSETVALUE:
            return self.__tacticRowBackground

    @tacticRowBackground.setter
    def tacticRowBackground(self, tacticRowBackground):
        typeChecker(type(self).__name__, tacticRowBackground, str,
                    "tacticRowBackground")
        self.__tacticRowBackground = tacticRowBackground

    @property
    def selectTechniquesAcrossTactics(self):
        if self.__selectTechniquesAcrossTactics != UNSETVALUE:
            return self.__selectTechniquesAcrossTactics

    @selectTechniquesAcrossTactics.setter
    def selectTechniquesAcrossTactics(self, selectTechniquesAcrossTactics):
        typeChecker(type(self).__name__, selectTechniquesAcrossTactics, bool,
                    "selectTechniqueAcrossTactics")
        self.__selectTechniquesAcrossTactics = selectTechniquesAcrossTactics

    @property
    def selectSubtechniquesWithParent(self):
        if self.__selectSubtechniquesWithParent != UNSETVALUE:
            return self.__selectSubtechniquesWithParent

    @selectSubtechniquesWithParent.setter
    def selectSubtechniquesWithParent(self, selectSubtechniquesWithParent):
        typeChecker(type(self).__name__, selectSubtechniquesWithParent, bool,
                    "selectSubtechniquesWithParent")
        self.__selectSubtechniquesWithParent = selectSubtechniquesWithParent

    @property
    def metadata(self):
        if self.__metadata != UNSETVALUE:
            return self.__metadata

    @metadata.setter
    def metadata(self, metadata):
        typeChecker(type(self).__name__, metadata, list, "metadata")
        self.__metadata = []

        for entry in metadata:
            try:
                if isinstance(entry, Metadata) or isinstance(entry, MetaDiv):
                    loadChecker(type(self).__name__, entry.get_dict(), ['name', 'value'], "metadata")
                    self.__metadata.append(entry)
                elif isinstance(entry, dict):
                    loadChecker(type(self).__name__, entry, ['name', 'value'], "metadata")
                    if entry['name'] == "DIVIDER":
                        self.__metadata.append(MetaDiv(active=entry['value']))
                    else:
                        self.__metadata.append(Metadata(name=entry['name'], value=entry['value']))
                else:
                    pass  # Object in the list was not of Metadata or MetaDiv type
            except MissingParameters as e:
                handler(
                    type(self).__name__,
                    'Metadata {} is missing parameters: {}. Skipping.'.format(entry, e)
                )

    @property
    def links(self):
        if self.__links != UNSETVALUE:
            return self.__links

    @links.setter
    def links(self, links):
        typeChecker(type(self).__name__, links, list, "links")
        if not handle_object_placement(self.__links, links, Link):
            self.__links = []
        entry = ""
        try:
            for entry in links:
                if isinstance(entry, Link):
                    loadChecker(type(self).__name__, entry.get_dict(), ['label', 'url'], "link")
                    self.__links.append(entry)
                elif isinstance(entry, LinkDiv):
                    loadChecker(type(self).__name__, entry.get_dict(), ['name', 'value'], "linkdiv")
                    self.__links.append(entry)
                elif isinstance(entry, dict):
                    if 'name' in entry and entry['name'] == "DIVIDER":
                        loadChecker(type(self).__name__, entry, ['name', 'value'], "linkdiv")
                        self.__links.append(LinkDiv(active=entry['value']))
                    else:
                        loadChecker(type(self).__name__, entry, ['label', 'url'], "link")
                        self.__links.append(Link(label=entry['label'], url=entry['url']))
                else:
                    pass
        except KeyError as e:
            handler(type(self).__name__, 'Link {} is missing parameters: '
                                         '{}. Unable to load.'
                    .format(entry, e))

    def _enumerate(self):
        """
            INTERNAL: Identifies which fields have been set for this Layer
                object
            :returns: a list of all set fields within this Layer object
        """
        temp = ['name', 'versions', 'domain']
        if self.description:
            temp.append('description')
        if self.filters:
            temp.append('filters')
        if self.sorting:
            temp.append('sorting')
        if self.layout:
            temp.append('layout')
        if self.hideDisabled:
            temp.append('hideDisabled')
        if self.techniques:
            temp.append('techniques')
        if self.gradient:
            temp.append('gradient')
        if self.legendItems:
            temp.append('legendItems')
        if self.showTacticRowBackground:
            temp.append('showTacticRowBackground')
        if self.tacticRowBackground:
            temp.append('tacticRowBackground')
        if self.selectTechniquesAcrossTactics:
            temp.append('selectTechniquesAcrossTactics')
        if self.selectSubtechniquesWithParent:
            temp.append('selectSubtechniquesWithParent')
        if self.metadata:
            temp.append('metadata')
        return temp

    def get_dict(self):
        """
            Converts the currently loaded layer into a dict
            :returns: A dict representation of the current layer object
        """
        temp = dict(name=self.name, domain=self.domain)

        if self.description:
            temp['description'] = self.description
        if self.versions:
            temp['versions'] = self.versions.get_dict()
        if self.filters:
            temp['filters'] = self.filters.get_dict()
        if self.sorting:
            temp['sorting'] = self.sorting
        if self.layout:
            temp['layout'] = self.layout.get_dict()
        if self.hideDisabled is not None:
            temp['hideDisabled'] = self.hideDisabled
        if self.techniques:
            temp['techniques'] = [x.get_dict() for x in self.techniques]
        if self.gradient:
            temp['gradient'] = self.gradient.get_dict()
        if self.legendItems:
            temp['legendItems'] = [x.get_dict() for x in self.legendItems]
        if self.showTacticRowBackground is not None:
            temp['showTacticRowBackground'] = self.showTacticRowBackground
        if self.tacticRowBackground:
            temp['tacticRowBackground'] = self.tacticRowBackground
        if self.selectTechniquesAcrossTactics is not None:
            temp['selectTechniquesAcrossTactics'] = \
                self.selectTechniquesAcrossTactics
        if self.selectSubtechniquesWithParent is not None:
            temp['selectSubtechniquesWithParent'] = \
                self.selectSubtechniquesWithParent
        if self.metadata:
            temp['metadata'] = [x.get_dict() for x in self.metadata]
        return temp

    def _linker(self, field, data):
        """
            INTERNAL: Acts as a middleman routing the settings of values
                within the layer
            :param field: The value field being set
            :param data: The corresponding data to set that field to
            :raises UnknownLayerProperty: An error indicating that an
                unexpected property was identified
        """
        if field == 'description':
            self.description = data
        elif field.startswith('version'):
            if not field.endswith('s'):
                # force upgrade
                print('[Version] - V3 version field detected. Upgrading to V4 Versions object.')
                ver_obj = dict(layer="4.0", navigator="4.0")
                self.versions = ver_obj
            else:
                self.versions = data
        elif field == 'filters':
            self.filters = data
        elif field == 'sorting':
            self.sorting = data
        elif field == 'layout':
            self.layout = data
        elif field == 'hideDisabled':
            self.hideDisabled = data
        elif field == 'techniques':
            self.techniques = data
        elif field == 'gradient':
            self.gradient = data
        elif field == 'legendItems':
            self.legendItems = data
        elif field == 'showTacticRowBackground':
            self.showTacticRowBackground = data
        elif field == 'tacticRowBackground':
            self.tacticRowBackground = data
        elif field == 'selectTechniquesAcrossTactics':
            self.selectTechniquesAcrossTactics = data
        elif field == 'selectSubtechniquesWithParent':
            self.selectSubtechniquesWithParent = data
        elif field == 'metadata':
            self.metadata = data
        elif field == 'links':
            self.links = data
        else:
            handler(type(self).__name__, "Unknown layer property: {}"
                    .format(field))
            raise UnknownLayerProperty



import argparse
import os

from mitreattack.navlayers.generators.overview_generator import OverviewLayerGenerator
from mitreattack.navlayers.generators.usage_generator import UsageLayerGenerator
from mitreattack.navlayers.generators.sum_generator import BatchGenerator


def main(argv=None):

    parser = argparse.ArgumentParser(description='Generate an ATT&CK Navigator layer')
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--overview-type', choices=['group', 'software', 'mitigation', 'datasource'],
                       help='Output a layer file where the target type is summarized across the entire dataset.')
    group.add_argument('--mapped-to', help='Output layer file with techniques mapped to the given group, software, '
                                           'mitigation, or data component. Argument can be name, associated '
                                           'group/software, or ATT&CK ID.')
    group.add_argument('--batch-type', choices=['group', 'software', 'mitigation', 'datasource'],
                       help='Output a collection of layer files to the specified folder, each one representing a '
                            'different instance of the target type.')
    parser.add_argument('-o', '--output', help='Path to the output layer file/directory', default='generated_layer')
    parser.add_argument('--domain', help='Which domain to build off of', choices=['enterprise', 'mobile', 'ics'],
                        default='enterprise')
    parser.add_argument('--source', choices=['taxii', 'local', 'remote'], default='taxii',
                        help='What source to utilize when building the layer files')
    parser.add_argument('--resource', help='Path to the local resource if --source=local, or url of an ATT&CK Workbench'
                                           ' instance if --source=remote', default=None)
    args = parser.parse_args(args=argv)

    if args.overview_type:
        og = OverviewLayerGenerator(source=args.source, domain=args.domain, resource=args.resource)
        generated = og.generate_layer(obj_type=args.overview_type)
        print('Generating Layer File')
        out_path = args.output
        if out_path == 'generated_layer':
            out_path += '.json'
        generated.to_file(out_path)
        print(f'Layer file generated as {out_path}.')
    elif args.mapped_to:
        ug = UsageLayerGenerator(source=args.source, domain=args.domain, resource=args.resource)
        generated = ug.generate_layer(match=args.mapped_to)
        print('Generating Layer File')
        out_path = args.output
        if out_path == 'generated_layer':
            out_path += '.json'
        generated.to_file(out_path)
        print(f'Layer file generated as {out_path}.')
    elif args.batch_type:
        bg = BatchGenerator(source=args.source, domain=args.domain, resource=args.resource)
        generated = bg.generate_layers(layers_type=args.batch_type)
        out_path = args.output
        if out_path == 'generated_layer':
            out_path += 's'
        if not os.path.exists(out_path):
            os.makedirs(out_path)
        for sid in generated:
            generated[sid].to_file(f"{out_path}/{sid}.json")
        print(f"Files saved to {out_path}/")


if __name__ == '__main__':
    main()
